#open "graphics";;

exception Erreur of string;;

let Pi = 4.*.atan(1.);;

(* la liste [1;...;n] *)
let range n =
	let rec range_int result = function
		|0 -> result
		|p -> range_int (p::result) (p-1)
in range_int [] n;;

(* le kime d'une liste *)
let rec kime k = function
	|t::q when k>0 -> if k=1 then t else kime (k-1) q
	|_ -> raise (Erreur "Argument non valide dans kime")
;;

(* Une fonction signe et d'arrondi *)
let sgn x = if x=0. then 0 else int_of_float (x/.(abs_float x));;

let trunc x = let E=float_of_int(int_of_float x) in
	if abs_float(x-.E)>.0.5 then (int_of_float E) +(sgn x)
					else int_of_float E
;;

(* Affichage d'une chane centre (??...) sur (x,y)				    .
Attention : x,y sont les coordonnes du centre suprieur du texte --> "Un exemple" *)
let affichestr x y str =
	let (dimx,dimy)=text_size str
in	begin
	moveto (x-dimx/2) y;
	draw_string str
	end
;;

(* Affichage d'un entier centr sur (x,y) *)
let afficheint x y n =
	let tmpStr = string_of_int n
in	let (dimx,dimy)=text_size tmpStr
in	begin
	moveto (x-dimx/2) y;
	draw_string tmpStr
	end
;;

(* Affichage d'une chane centre sur (x,y) *)
let print_str x y str =
	let (dimx,dimy)=text_size str
in	begin
	moveto (x-dimx/2) (y-dimy/2);
	draw_string str
	end
;;

(* Les courbes de Bzier (chantillone avec 200 points) *)
type PointOrVector= PoV of (float*float);;
type BezierCurve= Bezier of PointOrVector*PointOrVector*PointOrVector*PointOrVector;;

let CValues nbPts= let Step=1./.nbPts in
	let rec rValues t l=
	   let t2=t*.t in 
		let t3=t2*.t in
		   let ct=1.-.t and ct2=1.-.2.*.t+.t2 and ct3=1.-.3.*.t+.3.*.t2-.t3
	in	if t>.1. then l
		else rValues (t+.Step) ((t3,3.*.t2*.ct,3.*.t*.ct2,ct3)::l)
in	rValues 0. []
;;

let tValues=CValues 100.;;	(* Nombre de points : peut-tre  modifier *)

let DrawBezierCurve (Bezier((PoV(x1,y1)),(PoV(x2,y2)),(PoV(x3,y3)),(PoV(x4,y4)))) couleur= 
	let PutPt (x,y,z,t)= 
		lineto (trunc(x1*.x+.x2*.y+.x3*.z+.x4*.t)) 
			 (trunc(y1*.x+.y2*.y+.y3*.z+.y4*.t))
in 	begin
	set_color couleur;
	moveto (trunc x1) (trunc y1);
	do_list PutPt tValues
	end
;;

(* Calcule les coordonnes du "milieu" d'une Bezier *)
let MilieuCurve (Bezier((PoV(x1,y1)),(PoV(x2,y2)),(PoV(x3,y3)),(PoV(x4,y4)))) =
	let (a,b)=text_size "a"
in	trunc(x1*.0.125+.x2*.0.375+.x3*.0.375+.x4*.0.125 -. (float_of_int a /.2.)),
	trunc(y1*.0.125+.y2*.0.375+.y3*.0.375+.y4*.0.125 -. (float_of_int a /.2.))
;;

(* Dessin d'arbre *)
type point= {coord: int*int; val: string ref};;
type ('a,'b) arbre=
	|nil
	|feuille of 'a
	|noeud of 'b*(('a,'b) arbre list)
;;

let racine = function (* Renvoie la racine d'un arbre de string *)
	|nil -> ""
	|feuille f -> f
	|noeud(x,_) -> x
;;

let pasy=30;;				(* Dfinition de l'espacement entre les niveaux de l'arbre *)

let rec dessineliste liste_ab x y w affichea afficheb = match liste_ab with
	|[]   -> ()
	|t::q -> (rdessine t x y w affichea afficheb;
		    dessineliste q (x+w) y w affichea afficheb)
and rdessine arbre x y w affichea afficheb  = match arbre with
	|nil -> ()
	|(feuille(val)) -> affichea (x+w/2) y val
	|(noeud(_,[])) -> raise (Erreur "Arbre malform")
	|(noeud(val,l)) -> let longl=list_length l in let new_w=w/longl in
					begin
					afficheb (x+w/2) y val;
					for t=0 to (longl-1) do
						moveto (x+w/2) (y-14);
						lineto (x+new_w/2+t*new_w) (y-pasy)
					done;
					dessineliste l x (y-pasy) (new_w) affichea afficheb
					end
;;
let dessine arbre affichea afficheb couleur = 
	set_color couleur;
	rdessine arbre 0 (size_y()) (size_x()) affichea afficheb
;;

(* Saisie d'un texte au point (x,y) (Entre pour valider ou Echap pour annuler) *)
let saisie_int x y prev_val = let k=ref ` ` and result=ref "" in
	begin
	set_color black;
	affichestr x y (make_string ((string_length prev_val)+1) ` `);
	while not ((int_of_char !k)=13 or (int_of_char !k)=27) do
		affichestr x y (make_string ((string_length !result)+1) ` `);
		affichestr x y (!result);
		while not (key_pressed()) do done;
		k:=read_key();
		match (int_of_char !k) with
			|8 -> let longs=string_length !result in
					if longs=0 then ()
					else result:=sub_string (!result) 0 (longs -1)
			|n when n<32 -> ()
			|_ -> result:=!result^(char_for_read !k)
	done;
	affichestr x y (make_string ((string_length !result)+1) ` `);
	if (int_of_char !k)=13 then !result else prev_val;
	end
;;

(* Saisie d'un texte dans un cadre au centre de l'cran (Entre pour valider ou Echap pour annuler) *)
let draw_rect x y w h = 
	moveto x y;
	lineto (x+w) y; 
	lineto (x+w) (y+h);
	lineto x (y+h);
	lineto x y
;;
let saisie texte = 
	let result=ref ""  and taille=fst (text_size texte) +10 
in	let x=(size_x()-taille)/2 and y=size_y()/2-30
in	let fond=get_image x y (taille+1) 61
in	begin
	set_color background;
	fill_rect x y (taille+1) 61;
	set_color black;
	draw_rect x y taille 60;
	affichestr (size_x()/2) (size_y()/2+20) texte;
	result:=saisie_int (size_x()/2) (size_y()/2-5) "";
	draw_image fond x y;
	!result
	end
;;
let saisie_ext() = saisie "Entrer la nouvelle valeur :";;
let saisie_num() = saisie "Entrer le numro de l'tat :";;
let saisie_car() = saisie "Entrer le caractre de transition :";;

(* Affichage et effacement d'un curseur au point (x,y). Il ne clignote pas faute de timer *)
let dess_curs x y = set_color red; moveto (x-5) (y-13); lineto (x+5) (y-13);;
let eff_curs x y  = set_color background; moveto (x-5) (y-13); lineto (x+5) (y-13);;

(* Trac d'une ligne  la souris  partir de (x_fix,y_fix) et on rend l'autre extrmit *)
let trace_ligne statut (x_fix,y_fix) =
	let x_cur=ref 0 and y_cur=ref 0
(*	and statut=ref { mouse_x=0; mouse_y=0; button=false; keypressed=false; key=`\000` }*)
in	begin
	statut:={ mouse_x= !statut.mouse_x; mouse_y= !statut.mouse_y; button=false; keypressed=false; key=`\000` };
	while not(!statut.button) do
		x_cur:=!statut.mouse_x; y_cur:=!statut.mouse_y;
		statut:=wait_next_event [Button_down; Mouse_motion];	
		set_color background;
		moveto x_fix y_fix; lineto !x_cur !y_cur;
		set_color black;
		moveto x_fix y_fix; lineto !statut.mouse_x !statut.mouse_y;
	done;
	set_color background;
	moveto x_fix y_fix; lineto !statut.mouse_x !statut.mouse_y;
	(!statut.mouse_x, !statut.mouse_y); 
	end
;;

(* Fonctions de dessin pour les automates *)
type etat= {numero : int; pt : PointOrVector };;
type transition= {debut : etat; fin : etat; car : char; courbe : BezierCurve};;

let R=15.;;		(* Rayon des cercles autour des tats *)
let vert=rgb 16 209 21;;

let PoV_of_pt (x,y) = PoV(float_of_int x,float_of_int y);;
let pt_of_PoV (PoV(x,y)) = (trunc x,trunc y);;

(* Coordonnes du point situ  la distance l du point2 sur la droite (point1, point2) *)
let trouve_pt (x1,y1) (x2,y2) l = let hyp=sqrt(float_of_int ((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2))) in
	if x1=x2 then (x2,y2-(trunc l))
	else  let cosinus=float_of_int (x2-x1)/.hyp and sinus=float_of_int (y2-y1)/.hyp
		in (x2+trunc(l*.cosinus), y2+trunc(l*.sinus))
;;

(* Dessine une flche oriente selon le segment (point1, point2) *)
let trace_flche (x1,y1) (x2,y2) = 
	let deltax= x2-x1 and deltay= y2-y1
	and (x,y)= trouve_pt (x1,y1) (x2,y2) (-.6.)
in	let (a,b)= trouve_pt (x-deltay,y+deltax) (x,y) 4.
	and (c,d)= trouve_pt (x-deltay,y+deltax) (x,y) (-.4.)
in	begin
	moveto x2 y2;
	lineto a b;
	moveto x2 y2;
	lineto c d
	end
;;

(* Dessine une flche et le segment (point1, point2) *)
let dessine_flche (x1,y1) (x2,y2) = moveto x1 y1; lineto x2 y2; trace_flche (x1,y1) (x2,y2);;

(* Dessine une flche au bout d'une Bezier *)
(* /!\ : "plus" car les Beziers sont traces  l'envers (paramtre dcroissant) *)
#uninfix "plus";;
let plus (a,b) (c,d) = (a+c,b+d);;
#infix "plus";;

let flche_Bezier (Bezier((PoV(x1,y1)),(PoV(x2,y2)),(PoV(x3,y3)),(PoV(x4,y4)))) =
	let a=trunc(0.015625*.x1 +. 0.140625*.x2 +. 0.421875*.x3 +. 0.421875*.x4),
		trunc(0.015625*.y1 +. 0.140625*.y2 +. 0.421875*.y3 +. 0.421875*.y4)
	and b=trunc(0.1875*.x1   +. 0.9375*.x2   +. 0.5625*.x3   -. 1.6875*.x4),
		trunc(0.1875*.y1   +. 0.9375*.y2   +. 0.5625*.y3   -. 1.6875*.y4)
	and c=trunc(0.421875*.x1 +. 0.421875*.x2 +. 0.140625*.x3 +. 0.015625*.x4),
		trunc(0.421875*.y1 +. 0.421875*.y2 +. 0.140625*.y3 +. 0.015625*.y4)
	and d=trunc(1.6875*.x1   -. 0.5625*.x2   -. 0.9375*.x3   -. 0.1875*.x4),
		trunc(1.6875*.y1   -. 0.5625*.y2   -. 0.9375*.y3   -. 0.1875*.y4)
in	(trace_flche (a plus b) a; trace_flche (c plus d) c)
;;

(* Dessine une transition *)
let trace_transition col1 col2 transition =
	let caract=transition.car and curve=transition.courbe
in	let (x,y)=MilieuCurve curve
in	begin
	DrawBezierCurve curve col1;
	flche_Bezier curve;
	set_color col2; moveto x y; draw_char caract;
	end
;;

(* Dessine un tat *)
let trace_etat couleur state =
	let number=string_of_int state.numero and (x,y)=pt_of_PoV state.pt
in	let (a,b)=text_size number
in	begin
	set_color background;	
	fill_circle x y (trunc R);
	set_color couleur;	
	draw_circle x y (trunc R);
	moveto (x-a/2) (y-b/2); draw_string number
	end
;;

(* Dessine un tat initial *)
let trace_init couleur state = let (xi,yi)=pt_of_PoV state.pt in
	set_color couleur;
	if (xi-60)>=0
	then dessine_flche (xi-60, yi) (xi- trunc R,yi)
	else dessine_flche (xi,yi+60) (xi,yi+ trunc R)
;;

(* Dessine un tat final *)
let trace_fina couleur state = let (xf,yf)=pt_of_PoV state.pt in
	set_color couleur;
	draw_circle xf yf (trunc R -2)
;;

let efface_tt = clear_graph;;
