(****** TIPE d'Informatique (2me anne) ******)
(* modifi le 18/03/98 *)



(** Transformation Automate  ---> Dessin **)
directory "";; (* A modifier selon le systme *)
#load "graphics.cmo";;
open Graphics;;
#load "utilit.cmo";;
open Utilit;;

type automate2 = {
			ordre : int;
			initial : int list; final : int list;
			instantane : (int*int) list;
			transition : ((int*char)*int) list}
;;

(* Des constantes pour le dessin *) 
let pas_angl=pi/.12. and pas_tgte=50.;;

(* La fonction qui dessinera tout en fonction des rsultats futurs *)
let dessine_tt liste_etat liste_transition liste_instant init final =
	open_graph "";
	List.iter (trace_transition black red) liste_transition;
	List.iter (trace_transition vert vert) liste_instant;
	List.iter (trace_etat black) liste_etat;
	List.iter (trace_init black) init;
	List.iter (trace_fina black) final;
;;

(* Des utilitaires pour : 
		- transformer les transitions instantanes en "normales" 
		- faire le travail inverse de la fonction prcdente 
		- associer en couple tous les lments d'une liste
		- chercher les transitions entre deux tats
		- chercher les transitions entre deux classes
		- dfinir l'cartement entre les tats spcifiques  une classe *)
let rec transforme = function
	|[] -> []
	|(a,b)::q -> ((a,'\000'),b)::transforme q
;;

let separe =
	let rec rsepare norm inst = function
		|[] -> norm,inst
		|t::q ->	
			if t.car='\000'
			then rsepare norm ({debut=t.debut; fin=t.fin; car='\248'; courbe=t.courbe}::inst) q
			else rsepare (t::norm) inst q
in rsepare [] [];;

let rec associe =
	let rec couple result a = function
		|[] -> result
		|t::q -> couple ((a,t)::result) a q
in 	function
		|[] -> []
		|t::q -> (couple [t,t] t q)@(associe q)
;;

let cherche_transit etat1 etat2 =
	let rec rfind result = function
		|[] -> result
		|(((e,_),f) as t)::q -> 
			if (etat1=e & etat2=f) or (etat1=f & etat2=e)
			then rfind (t::result) q
			else rfind result q
in rfind []
;;

let cherche_passage classe1 classe2 =
	let rec rfind result = function
		|[] -> result
		|(((e,_),f) as t)::q ->
			if (List.mem e classe1 & List.mem f classe2) 
			or (List.mem e classe2 & List.mem f classe1)
			then rfind (t::result) q
			else rfind result q
in rfind []
;;

let def_step l = l,float_of_int (size_x()/(List.length l));;

(* La fonction finale *)
let dessine automate =
	let nbre=automate.ordre
	and liste_tot = automate.transition @ (transforme automate.instantane)

	(* La matrice de voisinage et le tableau qu va contenir les tats aprs leurs calculs *) 
in	let voisinage=make_matrix nbre nbre 0 and etats=Array.make nbre {numero=0; pt=PoV(0.,0.)}

	(* Initialisation de la matrice *) 
in	let cree_matrice () = 
		let rec construit= function
			|[] -> ()
			|((e,_),f)::q ->
				begin
				voisinage.(e-1).(f-1) <- voisinage.(e-1).(f-1) +1;
				voisinage.(f-1).(e-1) <- voisinage.(e-1).(f-1);
				construit q;
				end
		in construit liste_tot
	(* Recherche des tats ayant plus de deux transitions avec un tat *) 
	and copains etat = let result=ref [] in
		for t=0 to (nbre-1) do
			if voisinage.(etat-1).(t)>1
			then (result:= (t+1)::(!result); voisinage.(t).(etat-1)<-0)
		done;
		!result

	(* Dfinition des classes ou "noyaux" *) 
in	let rec def_classe () =
		let rec rdef result a_voir = match a_voir with
			|[] -> result
			|t::q ->	let voisins=copains t
					in let new= List.fold_left union [t] (rdef [] voisins)
					in rdef (new::result) (subtract a_voir new)
	in List.map def_step (rdef [] (range nbre))

	(* Calcul de la position des tats *) 
	and calcul_etats classes =
		let pas=float_of_int (size_y()/(List.length classes))
	in	let rec toto ord =
			let calcul (classe,step) = 
				let rec rcalc absc = function
					|[] -> ()
					|t::q ->	begin
							etats.(t-1) <- {numero=t; pt=PoV(absc,ord)};
							rcalc (absc+.step) q;
							end
			in rcalc (step/.2.) classe
		in function
			|[] -> ()
			|t::q -> (calcul t; toto (ord+.pas) q)
	in toto (pas/.2.) classes

	(* Calcul des transtions *) 
	and calcul_transitions =
		let rec calcul step = 
			let rec rcalc result (angle,tgte) fact =
				let calc ((e,c),f) =
					let ({numero=_; pt=PoV(x1,y1)} as etat1)=etats.(e-1)
					and ({numero=_; pt=PoV(x2,y2)} as etat2)=etats.(f-1) 
				in	let angl= if e=f 
								 then 4.*.pas_angl
								 else if x1<.x2 then angle else (Pi -. angle)
				in	let pt1= PoV(x1+.(r+.tgte)*.(cos angl),y1+.(r+.tgte)*.(sin angl))
					and pt2= PoV(x2-.(r+.tgte)*.(cos angl),y2+.(r+.tgte)*.(sin angl))
				in	{debut= etat1; fin= etat2; car=c;
					 courbe = Bezier (pov_of_pt (trouve_pt (pt_of_pov pt1) (pt_of_pov etat1.pt) (-.r)),
								pt1, pt2,
								pov_of_pt (trouve_pt (pt_of_pov pt2) (pt_of_pov etat2.pt) (-.r)))}
				and incr_data (t,l) = match t with
					|t when t>0. -> -.t,l
					|t when t<0. -> -.t +. (cos (-.t))/.3.,l+.pas_tgte
					|_ -> pas_angl*.fact,l
			in function
				|[] -> result
				|t::q -> rcalc (calc t::result) (incr_data (angle,tgte)) fact q
		in function
			|[] -> []
			|(e1,e2)::q->let (PoV(x1,y1))=(etats.(e1-1)).pt and (PoV(x2,y2))=(etats.(e2-1)).pt
					 in let mult= float_of_int (trunc((abs_float (x1-.x2)) /. step)-1)
					 in (rcalc [] (0.+. 2.*.mult*.pas_angl,pas_tgte) (mult+.1.) (cherche_transit e1 e2 liste_tot))
					    @(calcul step q)

	in function
		|[] -> []
		|(classe,step)::q -> (calcul step (associe classe))@(calcul_transitions q)

	(* Calcul de des transitions entre les classes *) 
	and calcul_passage classes = let pas=float_of_int (size_y()/(List.length classes)) in
		let rec calcul =
			let rec rcalc result =
				let calc ((e,c),f) =
					let ({numero=_; pt=PoV(x1,y1)} as etat1)=etats.(e-1)
					and ({numero=_; pt=PoV(x2,y2)} as etat2)=etats.(f-1) 
				in	let tang = pas_tgte *. (float_of_int (sgn (y2-.y1)))
					and ray = r *. (float_of_int (sgn (y2-.y1)))
					and diff=abs_float (y2-.y1)/.pas -.1.
				in	let pt1 = if x1=x2
								 then PoV(x1+.40.*.diff, y1+.ray+.tang)
								 else PoV(x1, y1+.ray+.tang) 
					and pt2 = if x1=x2
								 then PoV(x2+.40.*.diff, y2-.ray-.tang)
								 else PoV(x2, y2-.ray-.tang)
				in	{debut= etat1; fin= etat2; car= c;
					 courbe = Bezier (PoV(x1, y1+.ray),pt1, pt2,PoV(x2, y2-.ray))}
			in function
				|[] -> result
				|t::q -> rcalc (calc t::result) q
		in function
			|[] -> []
			|((c1,_),(c2,_))::q -> (rcalc [] (cherche_passage c1 c2 liste_tot))@(calcul q)
	in calcul (subtract (associe classes) (List.map (function x -> x,x) classes))

	(* Applications de tout ce qui prcde ... *)
in	let classes=(cree_matrice(); def_classe()) in
		let liste_etats=(calcul_etats classes; List.of_vect etats)
		and (liste_transition,liste_instant)=separe ((calcul_transitions classes)@(calcul_passage classes))
		and etat_init=List.map (function n -> etats.(n-1)) automate.initial
		and etat_final=List.map (function n -> etats.(n-1)) automate.final
			in dessine_tt liste_etats liste_transition liste_instant etat_init etat_final
;;

(* Quelques exemples ... *) 
(* Un qui marche bien avec deux classes *) 
let automate = 
 {ordre=4; initial = [1]; final = [4]; instantane = [3, 2; 1, 3];
  transition =
   [(3, 'a'), 4; (3, 'c'), 1; (3, 'b'), 2; (3, 'a'), 2; (1, 'b'), 3;
    (1, 'c'), 2; (2, 'b'), 1; (1, 'a'), 2; (2, 'd'), 2]};;
dessine automate;;
read_key();;

(* Un qui marche pas bien (sans classes) *) 
let automate = 
 {ordre=4; initial = [1]; final = [4]; instantane = [3, 2; 1, 3; 2, 4];
  transition =
   [(3, 'a'), 4; (3, 'c'), 1; (3, 'b'), 2; (3, 'a'), 2; (1, 'b'), 3;
    (4, 'b'), 3; (2, 'b'), 1; (1, 'a'), 2; (2, 'f'), 2;
    (2, 'f'), 3; (1, 'a'), 2; (1, 'c'), 4; (4, 'e'), 2;
    (1, 'c'), 2; (2, 'b'), 1; (1, 'a'), 2; (2, 'd'), 2]};;
dessine automate;;
read_key();;

(* Le tonneau du diable *) 
let automate =
 {ordre = 3; initial = [1]; final = [1]; instantane = [];
  transition =
   [(2, 'h'), 2; (2, 'b'), 2; (2, 'm'), 1; (1, 'd'), 2; (2, 'd'), 3;
    (1, 'm'), 3; (1, 'h'), 1; (1, 'b'), 1]};;
dessine automate;;
read_key();;

(* ENSI '97 (marche pas bien !!) *) 
let automate =
 {ordre = 4; initial = [1]; final = [4]; instantane = [];
  transition =
   [(3, 'a'), 3; (3, 'b'), 3; (2, 'a'), 2; (2, 'b'), 4; (4, 'a'), 3;
    (4, 'b'), 1; (1, 'b'), 3; (1, 'a'), 2]};;
dessine automate;;
read_key();;

(* Exemple du rapport TIPE *) 
let automate =
 {ordre = 3; initial = [1]; final = [3]; instantane = [2, 1];
  transition =
   [(3, 'a'), 3; (2, 'b'), 2; (3, 'b'), 2; (2, 'a'), 3; (1, 'b'), 2;
    (1, 'a'), 3]};;
dessine automate;;
read_key();;

(* ENSI '98 (relativement correct) *) 
let automate =
 {ordre = 8; initial = [1]; final = [4]; instantane = [];
  transition =
   [(8, 'b'), 8; (8, 'a'), 8; (5, 'b'), 8; (5, 'a'), 4; (4, 'b'), 5;
    (7, 'b'), 7; (7, 'a'), 7; (4, 'a'), 7; (3, 'a'), 4; (2, 'a'), 4;
    (6, 'b'), 6; (6, 'a'), 6; (3, 'b'), 6; (2, 'b'), 6; (1, 'b'), 3;
    (1, 'a'), 2]};;
dessine automate;;
read_key();;

(* Mines '98 (marche bien...)*) 
let automate =
 {ordre = 3; initial = [1]; final = [3]; instantane = [];
  transition =
   [(3, 'b'), 2; (2, 'b'), 3; (2, 'a'), 3; (3, 'a'), 1; (1, 'a'), 2;
    (1, 'b'), 1]};;
dessine automate;;
read_key();;

(* Centrale '98 (NUL et illisible !!)*) 
let automate =
{ordre = 14; initial = [1]; final = [14]; instantane = [];
  transition =
   [(13, 'a'), 14; (11, 'a'), 13; (10, 'a'), 12; (9, 'b'), 13; (12, 'b'), 14;
    (8, 'b'), 12; (7, 'c'), 11; (6, 'c'), 10; (5, 'c'), 9; (4, 'c'), 8;
    (3, 'b'), 6; (2, 'b'), 4; (1, 'b'), 2; (2, 'a'), 5; (3, 'a'), 7;
    (1, 'a'), 3]};;
dessine automate;;
read_key();;
