(****** TIPE d'Informatique (2me anne) ******)
(* modifi le 16/02/98 *)



(** Transformation Dessin ---> Automate **)
directory "";; (* A modifier selon le systme *)
#load "graphics.cmo";;
open Graphics;;
#load "utilit.cmo";;
open Utilit;;

type automate1 = {
	ordre : int;
	initial : int; final : int -> bool;
	transition : (int*char) -> int
};;

(* La fonction finale de dessin *)
let dessine_tt liste_etat liste_transition init final =
	begin
	List.iter (trace_transition black red) liste_transition;
	List.iter (trace_etat black) liste_etat;
	trace_init black init;
	trace_fina black final;
	end
;;

(* Les distance et norme euclidiennes *)
let norme (PoV(x,y)) = sqrt(x*.x +. y*.y);;
let dist (PoV(x1,y1)) (PoV(x2,y2)) = norme (PoV(x1-.x2,y1-.y2));;

(* Calcul de l'tat le plus proche d'un autre tat *) 
let etat_proche pt_fix =
	let rec min_dist voisin mini = function
		|[]	-> voisin
		|t::q -> let n=dist t.pt pt_fix in 
				if mini<.n
				then min_dist voisin mini q
				else min_dist t n q
in	min_dist {numero= -1; pt=PoV(-10.,-10.)} (float_of_int(size_x()*size_x()))
;;

(* Remplace point par sa nouvelle valeur dans une liste *)
let rec remplace etat = function
	|[]	-> raise (Erreur "Erreur dans remplace : Etat non prsent")
	|t::q -> if  t.pt=etat.pt then etat::q else t::(remplace etat q)
;;

(* Supprime etat de la liste des tats et des transitions *)
let rec supprime_ds_etat etat = function
	|[]	-> raise (Erreur "Erreur dans supprime_ds_etat : Etat non prsent")
	|t::q -> if  t.pt=etat.pt then q else t::(supprime_ds_etat etat q)
;;
let rec supprime_ds_transition etat = function
	|[]	-> []
	|t::q -> if (t.debut=etat or t.fin=etat)
				then supprime_ds_transition etat q
				else t::(supprime_ds_transition etat q)
;;

(* Supprime de la liste des transitions la transition qui a pour "milieu" (x,y) *)
let rec supprime (x,y) = function
	|[]	-> raise (Erreur "Erreur dans supprime : Transition non prsente")
	|t::q -> let (a,b)=milieuCurve t.courbe in 
			if (abs(a-x)<=7 & abs (b-y)<=18)
			then q 
			else t::(supprime (x,y) q)
;;

(* L'interface de dessin  :
	La fonction dessin va retourner, pour le traitement, la liste des tats
et des transitions cres, ainsi que les tats initial et final du futur 
automate. Elle utilise en grande partie la souris pour la cration de nouveaux 
tats, leur dplacement et la slection des tats final et initial. 
	Pour dfinir une transition, il faut appuyer sur 'T' ou 't' puis de 
choisir les 4 points de contrde la Bezier. Si vous terminez  un endroit 
o il n'y a pas d'tat, un nouvel tat est cr. On peut aussi supprimer un 
tat en appuyant sur 'S' ou 's' puis en cliquant sur l'tat, ou supprimer une 
transition en appuyant sur 'D' ou 'd' (comme Delete ...) puis en cliquant sur 
son caractre. Toutes les transitions partant de l'tat supprim point seront 
alors supprimes. 
	Pour sortir, appuyer sur 'Echap'. Il faut avoir pralablement choisi un tat 
final et un tat initial en appuyant sur 'F' ou 'f', respectivement sur 'I' ou 
'i', puis en cliquant sur l'tat appropri. Les transitions non prsentes seront 
considres comme tombant dans un puit. *)

let dessin () =
	let statut=ref {mouse_x=0; mouse_y=0; button=false; keypressed=false; key='\000'}

	and temp=ref {numero= -1; pt=PoV(-10.,-10.)}
	and etat_init=ref {numero= -1; pt=PoV(-10.,-10.)}
	and etat_fina=ref {numero= -1; pt=PoV(-10.,-10.)}
	and liste_etat= ref [] and liste_transition=ref []

	(* Dit si pt2 est dans le cercle de pt1 *) 
in	let est_ds_pt pt1 pt2 = (dist pt1 pt2)<=r

	(* Traitement des nouveaux tats (ou dj prsents !!) *) 
in	let traite_newetat etat_ref point = let state=etat_proche (pov_of_pt point) !liste_etat in
		if (est_ds_pt state.pt (pov_of_pt point))
		then etat_ref:=state
		else let new_num=int_of_string (saisie_num()) in 
			(etat_ref:={numero=new_num; pt=pov_of_pt point};
			 liste_etat:=!etat_ref::!liste_etat)

	and attend_pt () = 
		begin
		statut:=wait_next_event [Button_down];
		(!statut.mouse_x,!statut.mouse_y)
		end

in 	let cree_transition () = 
		let deb=ref {numero=0; pt=PoV(0.,0.)} and fini=ref {numero=0; pt=PoV(0.,0.)}
		and pt1=ref (0,0) and pt2=ref (0,0)
	in	begin
		traite_newetat deb (attend_pt());
		trace_etat black !deb;
		if deb=etat_fina then trace_fina black !etat_fina;
		pt1:=attend_pt(); pt2:=attend_pt();
		traite_newetat fini (attend_pt());
		liste_transition:={debut= !deb; fin= !fini; car=(saisie_car()).[0];
					 courbe = Bezier (pov_of_pt (trouve_pt !pt1 (pt_of_pov !deb.pt) (-.r)),
								pov_of_pt (trouve_pt (pt_of_pov !deb.pt) !pt1 r),
								pov_of_pt (trouve_pt (pt_of_pov !fini.pt) !pt2 r),
								pov_of_pt (trouve_pt !pt2 (pt_of_pov !fini.pt) (-.r)))}
		::!liste_transition
		end
								

in	while (Char.code !statut.key)<>27 do
		dessine_tt !liste_etat !liste_transition !etat_init !etat_fina;
		statut:=wait_next_event [Button_down; Key_pressed];

		if !statut.button then traite_newetat temp (!statut.mouse_x,!statut.mouse_y)
		else 	if !statut.keypressed then
				begin
				match (Char.code !statut.key) with
(* F *)				|70|102 -> traite_newetat etat_fina (attend_pt())

(* I *)				|73|105 -> traite_newetat etat_init (attend_pt())

(* T *)				|84|116 -> cree_transition()

(* S *)				|83|115 ->
							begin
							traite_newetat temp (attend_pt());
							efface_tt();
							liste_etat:=supprime_ds_etat !temp !liste_etat;
							liste_transition:=supprime_ds_transition !temp !liste_transition;
							end

(* D *)				|68|100 -> let milieu=attend_pt() in
							begin
							efface_tt();
							liste_transition:=supprime milieu !liste_transition;
							end

					|_ -> ()
				end
			else ()
	done;
	(!liste_transition,!etat_init,!etat_fina,List.length !liste_etat)
;;

(* open_graph ""; dessin();; *)

(* La fonction finale *)
let boulot () = let (transit, init, fina, nbre)= (open_graph ""; dessin())
in	let rec liste_a_fonction liste =
		let intermediaire (e,a) = 
			try (List.assoc (e,a) liste) 
			with Not_found -> failwith "Transition non prvue"
		in intermediaire
	and transforme = function
		|[]	-> []
		|t::q -> ((t.debut.numero,t.car),t.fin.numero)::transforme q
in	{ordre=nbre;
	 initial=init.numero; final=(function n -> n=fina.numero);
	 transition=liste_a_fonction (transforme transit)}
;;

(*boulot();;*)


