(* Types *)
type point = Pt of (float*float);;
type point_pond = Ptm of point*float;;

(* Lois de composition internes et externes sur les points *)
let add (Pt(a,b)) (Pt(c,d)) = Pt(a+.c, b+.d);;
let sub (Pt(a,b)) (Pt(c,d)) = Pt(a-.c, b-.d);;
let mult lambda (Pt(a,b)) = Pt(lambda*.a, lambda*.b);;
let divi (Pt(a,b)) lambda = Pt(a/.lambda, b/.lambda);;

(* Calcul de distance euclidienne *)
let dist (Pt(a,b)) (Pt(c,d)) = sqrt((a-.c)*.(a-.c)+.(b-.d)*.(b-.d));;

(* Fonctions "pousseuses" *)
let deplacement masse ratio =
	if ratio <=. 0.5
	then 4./.(masse+.1.)*.ratio*.(1.-.ratio)*.masse
	else 1.+.4./.(masse+.1.)*.ratio*.(ratio-.1.)
;;
let deplacement masse t =
	(1.-.t)*.(power t (1./.(sqrt masse)))+.t*.t;
;;

(* Recherche du point de la frontire situ sur la droite (pt_o, pt_f) *)
let pt_front (Pt(x_o,y_o)) (Pt(x_f,y_f) as pt_f) frontire =
	let demi_droite (Pt(x,y)) = x*.(y_o-.y_f) -. y*.(x_o-.x_f) +. x_o*.y_f -. y_o*.x_f

	and pt_inter (Pt(x_a,y_a),Pt(x_b,y_b)) =
		let a=y_o-.y_f and b=x_f-.x_o and c=x_o*.y_f-.y_o*.x_f
		and u=y_a-.y_b and v=x_b-.x_a and w=x_a*.y_b-.y_a*.x_b
		in let det=a*.v-.b*.u
		in if (u,v)=(0.,0.)
			then (Pt(x_a,y_a))
			else divi (Pt(w*.b-.v*.c,c*.u-.a*.w)) det

	and traite_pt_front old (Pt(x,y) as new) =
			if ((x-.x_o)*.(x_f-.x_o)>.0.
			 &  (y-.y_o)*.(y_f-.y_o)>.0.
			 &  (dist pt_f old)>.(dist pt_f new))
			then new
			else old

in	let rec cherche result prec = function
			| [] -> result
			| t::q ->
					if demi_droite t=0. 
					then cherche (traite_pt_front result t) t q
					else if (demi_droite t)*.(demi_droite prec) <.0.
				   	  then cherche (traite_pt_front result (pt_inter (prec,t))) t q
				   	  else cherche result t q

in	cherche (Pt(790200.,2143200.)) (hd frontire) frontire;;

(* Dplacement d'un point *)
let deplace_pt frontiere list_ptm long (Pt(x,y) as pt1) =
	let f (Ptm(pt2,m)) = 
		let pt_lim = pt_front pt2 pt1 frontiere
	in	let ratio = (dist pt2 pt1)/.(dist pt2 pt_lim)
		and v_dir=sub pt_lim pt2
	in	let move=deplacement m ratio in
			add pt2 (mult move v_dir)
in let calc pt1 pt2 = add pt1 (divi (f pt2) long)
in it_list calc (Pt(0.,0.)) list_ptm;;

(* Dplacement de tous les points *)
let deplace_tt liste_pt frontiere liste_ptm =
	map (deplace_pt frontiere liste_ptm (float_of_int (list_length liste_ptm))) liste_pt;;
