(** Moulinette **)


(* Initialisation *)
directory "A:";;
directory "D:\Program Files\Caml\Fichiers\TIPE";;

#open "Deplace";; load_object "Deplace";;
#open "genlex";;
#open "sys";;
#open "graphics";;
(* #open "graphx";; load_object "graphx";; *)

(* Types *)
type commune =
	 { Nom : string;
		Population : int;
		Tx_vacance : float;
		mutable Contour : (point list list);
		Centre : point
	 };;

(* Utilitaires *)
let rec stream_of_list = function
	| [] -> [< >]
	|t::q -> [<'t; stream_of_list q >]
;;

let rec list_of_stream = function
	| [< 't; list_of_stream q >] -> t::q
	| [< >] -> []
;;

let partie m liste =
   let rec take result = function
      | (0,_) -> result
      | (_,[]) -> failwith "Partie trop petite"
      | (k,t::q) -> take (t::result) (k-1,q)
in take [] (m,liste);;

#open "random";;
let partie_alea ensemble =
   let rec prend_kieme_et_enlve ens = function
      | 1 -> (hd ens, tl ens)
      | k -> let (elt,reste)=prend_kieme_et_enlve (tl ens) (k-1)
               in (elt, (hd ens)::reste)
in let rec cree result ens n = function
      | 0 -> result
      | p -> let (new,reste)=prend_kieme_et_enlve ens (int n +1)
               in cree (new::result) reste (n-1) (p-1)
in cree [] ensemble (list_length ensemble);;

(* Lecture des fichiers de donnes *)
let lecture fichier =
	let canal1=open_in (fichier^".mid") and canal2=open_in (fichier^".mif")
in let flux1=make_lexer [","] (stream_of_channel canal1)
	and flux2=make_lexer ["Region";"Center"] (stream_of_channel canal2)

in let lie_nb_regions = function
		| [< 'Kwd "Region"; 'Int n >] -> n
		| [< >] -> failwith "Erreur lie_nb_regions"

	and lie_centre = function
		| [< 'Kwd "Center"; 'Int x; 'Int y >] -> Pt(float_of_int x,float_of_int y)
		| [< >] -> failwith "Erreur lie_centre"

in let rec lie_points p =
		if p=0
		then function [< >] -> []
		else function
			[< 'Int x; 'Int y; (lie_points (p-1)) liste >] -> (Pt(float_of_int x,float_of_int y))::liste
	
	and lie_regions n = 
		if n=0
		then function [< >] -> []
		else function
			[< 'Int p; (lie_points p) liste_pt; (lie_regions (n-1)) liste_tot >] -> liste_pt::liste_tot

in let rec lie = function
		| [< '_;'Kwd _;'String n;'Kwd _;'_;'Kwd _;'Int p;'Kwd _;'Float t; lie temp >] -> 
			let nb=lie_nb_regions flux2 in
			let reg=lie_regions nb flux2 in
			let cen=lie_centre flux2 in
				{Nom=n; Population=p; Tx_vacance=t; Contour=reg; Centre=cen}::temp
		| [< >] -> []

in let result=lie flux1
in (close_in canal1; close_in canal2; result);;

let rec lie_frontiere fichier =
	let canal=open_in fichier
in let rec lie = function
			|[< 'Int x; 'Int y; lie temp >] -> (Pt(float_of_int x,float_of_int y))::temp
			|[< >] -> []
in let result=lie (make_lexer [] (stream_of_channel canal))
in (close_in canal; result);;

(* Recherche du plus grand rectangle contenant la carte *)
let cherche_min_max =
	let rec min_max (mini_x,mini_y,maxi_x,maxi_y) = function
		| [] -> (mini_x,mini_y,maxi_x,maxi_y)
		| (Pt(x,y))::q -> min_max (min x mini_x,min y mini_y,max x maxi_x,max y maxi_y) q
	and parcourt rect_max = function
		| [] -> rect_max
		| t::q -> let new=it_list min_max rect_max t.Contour in parcourt new q
in parcourt (float_of_int max_int,float_of_int max_int,0.,0.);;

(* Recherche des valeurs min et max de la donne *)
let cherche_min_max_val =
	let rec parcourt mini maxi = function
		| [] -> (mini,maxi)
		| t::q -> parcourt (min t.Population mini) (max t.Population maxi) q
in parcourt 1000000 0;;

(* Fonctions de trac *)
let min_x=705500. and min_y=2027900. and max_x=790100. and max_y=2143100.;;
let min_pop=55. and max_pop=199528.;;
(* set_line_width 0; set_font "Courier"; set_font_size 5; *)

let echelle (Pt(x,y)) =
	let homoth=min (float_of_int (size_x())/.(max_x-.min_x)) (float_of_int (size_y())/.(max_y-.min_y))
in Pt((x-.min_x)*.homoth,(y-.min_y)*.homoth)
;;

let ligne pt = let (Pt(x,y))=echelle pt in lineto (int_of_float x) (int_of_float y)
and bouge pt = let (Pt(x,y))=echelle pt in moveto (int_of_float x) (int_of_float y);;
let print_str pt str =
	let (dimx,dimy)=text_size str and (Pt(x,y))=echelle pt
in	begin
	moveto (int_of_float (x -. float_of_int dimx /.2.)) (int_of_float (y -. float_of_int dimy /.2.));
	draw_string str
	end
;;

let trace_contour commune =
	let dessine l = (bouge (hd l); do_list ligne (tl l))
in do_list dessine commune.Contour;
	print_str commune.Centre commune.Nom
;;

(* Permet d'obtenir la carte avec le gradient de couleur pour la population *)
(*
let ligne pt = let (Pt(x,y))=echelle pt in lineto {x=int_of_float x; y=int_of_float y}
and bouge pt = let (Pt(x,y))=echelle pt in moveto {x=int_of_float x; y=int_of_float y};;
let print_str pt str =
	let (dimx,dimy)=text_size str and (Pt(x,y))=echelle pt
in	begin
	moveto {x=(int_of_float (x -. float_of_int dimx /.4.));y=(int_of_float (y -. float_of_int dimy /.4.))};
	draw_string str
	end
;;

let trace_contour commune =
	let f pt=let (Pt(x,y))=echelle pt in (int_of_float x,int_of_float y)
	and calc_col pop = let temp=float_of_int pop in
		let toto=255- int_of_float ((power ((temp-.min_pop)/.(max_pop-.min_pop)) (1./.3.))*.255.)
		in rgb toto toto toto
in	let dessine l = fill_poly (vect_of_list (map f l))
in	set_color (calc_col commune.Population);
	do_list dessine commune.Contour;
;;

let trace_contour commune =
	let f pt=let (Pt(x,y))=echelle pt in {x=int_of_float x; y=int_of_float y}
	and calc_col pop = let temp=float_of_int pop in
		let toto=255- int_of_float ((power ((temp-.min_pop)/.(max_pop-.min_pop)) (1./.3.))*.255.)
		in rgb toto toto toto
in	let dessine l = fill_poly (vect_of_list (map f l))
in	set_color (calc_col commune.Population);
	do_list dessine commune.Contour;
;;
*)

let trace_carte = do_list trace_contour;;

let trace_front l = 
	bouge (hd l); do_list ligne (tl l)
;;

(* Essai de trac *)
open_graph "";
set_color red; trace_front (lie_frontiere "D:\Program Files\Caml\Fichiers\TIPE\lim42.mif");
set_color black; trace_carte (lecture "D:\Program Files\Caml\Fichiers\TIPE\Loire");;

(* Trace une carte  partir d'un fichier sans le charger *)
let trace_file fichier =
	let canal=open_in fichier
in let flux=make_lexer ["Region";"Center"] (stream_of_channel canal)

in let lie_nb_regions = function
		| [< 'Kwd "Region"; 'Int n >] -> n
(*		| [< >] -> failwith "Erreur lie_nb_regions"*)

	and lie_centre = function
		| [< 'Kwd "Center"; 'Int x; 'Int y >] -> ()
		| [< >] -> failwith "Erreur lie_centre"

in let rec trace_points p fin =
		if p=fin
		then function [< 'Int x; 'Int y >] -> bouge (Pt(float_of_int x,float_of_int y))
		else function
			[< 'Int x; 'Int y; (trace_points (p+1) fin) () >] -> ligne (Pt(float_of_int x,float_of_int y))
	
	and trace_regions n = 
		if n=0
		then function [< >] -> ()
		else function
			[< 'Int fin; (trace_points 1 fin) (); (trace_regions (n-1)) () >] -> ()

in let rec dessine () =
	try let nb=lie_nb_regions flux in
		 	begin
		 	trace_regions nb flux;
		 	lie_centre flux;
		 	dessine ()
			end
	with Parse_error-> ()
	    |Parse_failure -> ()

in (dessine (); close_in canal);;

(* Fonctions d'extraction du rsultat *)
let rec extrait = function
	| [] -> ([],[])
	|t::q -> let (a,b)=extrait q in
		((it_list (prefix @) [] t.Contour)@a,(Ptm(t.Centre,float_of_int t.Population))::b)
;;

let rec prend k =
	if k=0
	then function [< >] -> []
	else function [< 't; (prend (k-1)) result >] -> t::result
;;

let rec remet new = 
	let rec change_ll = function
		| [] -> []
		|t::q -> let a=prend (list_length t) new in a::(change_ll q)
in function
	| [] -> ()
	| t::q -> (t.Contour<-(change_ll t.Contour); remet new q)
;;

(* Fonctions de sauvegarde du rsultat *)
let sauve fichier flux =
	let canal=open_out_gen [O_RDWR; O_CREAT; O_TRUNC; O_TEXT] s_iwusr fichier
in	let rec boulot = function
		| [< 'Pt(x,y); boulot () >] -> 
			let chaine=(string_of_float x)^" "^(string_of_float y)^"\n"
			in output_string canal chaine
		| [< >] -> ()
in (boulot flux; close_out canal);;

let ecrit fichier liste =
	let canal=open_out_gen [O_RDWR; O_CREAT; O_TRUNC; O_TEXT] s_iwusr fichier
in	let write_pt (Pt(x,y)) = 
			let chaine=(string_of_int (int_of_float x))^" "^(string_of_int (int_of_float y))^"\n"
			in output_string canal chaine

in let rec ecrit_pt = function
		| [] -> ()
		| t::q -> (write_pt t; ecrit_pt q)
	and ecrit_cont l_pt =
		begin
		output_string canal (" "^(string_of_int (list_length l_pt))^"\n");
		ecrit_pt l_pt;
		end

in let rec boulot = function
		| [] -> ()
		|t::q -> 
			begin
			output_string canal ("Region "^(string_of_int (list_length t.Contour))^"\n");
			do_list ecrit_cont t.Contour;
			output_string canal " Center "; write_pt t.Centre;
			boulot q
			end

in (boulot liste; close_out canal);;

(* Essai de calcul (/!\ demande environ 4h30 de calcul !!)*)
let temp=lecture "D:\Program Files\Caml\Fichiers\TIPE\Loire";;
let (liste_pt,liste_ptm)=extrait temp
and frontiere=lie_frontiere "D:\Program Files\Caml\Fichiers\TIPE\lim42.mif";;

let t=time();;
let toto=deplace_tt liste_pt frontiere liste_ptm;;
print_float (time()-.t);;
sauve "D:/Program Files/Caml/Fichiers/TIPE/temp.txt" (stream_of_list toto);;
remet (stream_of_list toto) temp;;
ecrit "D:/Program Files/Caml/Fichiers/TIPE/Loire2.mif" temp;;

(* Trac du rsultat final *)
let gray=rgb 200 200 200;;
open_graph "";
set_color gray	; trace_file "D:/Program Files/Caml/Fichiers/TIPE/Loire.mif";
set_color black; trace_file "D:/Program Files/Caml/Fichiers/TIPE/Loire2.mif";;
(* write_postscript "Carte2.ps";; *)

(* Evaluation du temps de calcul *)
let pt1=hd (tl liste_pt) and (Ptm(pt2,_))=hd liste_ptm in
pt_front pt1 pt2 frontiere;;

let test1 () =
	let pt1=hd liste_pt and t=ref 0. and a=ref (Pt(0.,0.))
	and long=float_of_int (list_length liste_ptm)
in	t:=time();
	a:=deplace_pt frontiere liste_ptm long pt1;
	print_float (time()-. !t);
	print_newline ();
(*	t:=time();
	a:=deplace_pt2 frontiere liste_ptm pt1;
	print_float (time()-. !t)*)
;;

let test2 k =
	let partiel=partie k liste_pt and t=ref 0. and a=ref []
in	t:=time();
	a:=deplace_tt partiel frontiere liste_ptm;
	print_float (time()-. !t);
;;

let test3 k =
	let partiel=partie_alea liste_pt k and t=ref 0. and a=ref []
in	t:=time();
	a:=deplace_tt partiel frontiere liste_ptm;
	print_float (time()-. !t);
;;
