(****** TIPE d'informatique ******)


#load "graphics.cmo";;
open Graphics;;

(* Dclaration des types *)
type pt3d = P3 of (float * float * float);;
type pt2d = P2 of (float * float);;
(* Exit les pointeurs : avec les index, mmes avantages et en plus la mme 
valeur convient pour espace et plan... *)
type face = F of (int * int * int * int * int * color);;

type pointori=POri of (float * float * float * int * int * int);;
(* x y z theta phi alpha ; pointori= point orient) *)

(* Dclaration des constantes *)
let f = 200.;;						(* Dfinition de la distance observateur/cran *)
let offx = float_of_int (size_x()) /. 2.;;	(* Milieu de l'cran *) 
let offy = float_of_int (size_y()) /. 2.;;

(* Fonctions d'accs de base  nos types *) 
let x=fun (P3(a,_,_))->a;; 
let y=fun (P3(_,a,_))->a;; 
let z=fun (P3(_,_,a))->a;;

let absc=fun (P2(a,_))->a;;
let ord=fun (P2(_,a))->a;;

let p1=fun (F(a,_,_,_,_,_))->a;;
let p2=fun (F(_,a,_,_,_,_))->a;;
let p3=fun (F(_,_,a,_,_,_))->a;; 
let p4=fun (F(_,_,_,a,_,_))->a;; 
let normale=fun (F(_,_,_,_,a,_))->a;;
let couleur=fun (F(_,_,_,_,_,a))->a;;

let xo=fun (POri(a,_,_,_,_,_))->a;;
let yo=fun (POri(_,a,_,_,_,_))->a;; 
let zo=fun (POri(_,_,a,_,_,_))->a;;
let theta=fun (POri(_,_,_,a,_,_))->a;; 
let phi=fun (POri(_,_,_,_,a,_))->a;; 
let alpha=fun (POri(_,_,_,_,_,a))->a;;

(* Oprations sur les vecteurs/points *)
let plus u v = P3((x u) +. (x v), (y u) +. (y v), (z u) +. (z v));;

let neg (P3(xu,yu,zu))= (P3(-.xu,-.yu,-.zu));;

let moins u v = plus u (neg v);;

let prod_scal_norm (P3(xa,ya,za)) (P3(xb,yb,zb)) = 
	(xa*.xb+.ya*.yb+.za*.zb)/.(sqrt((xa*.xa+.ya*.ya+.za*.za)*.(xb*.xb+.yb*.yb+.zb*.zb)));;

(* Projection conique valable pr tt objet situ  x>0 *)
let proj = fun (P3(a,b,c))->(P2(b *. f /. a +. offx, c *. f /. a +. offy));;

(* Affichage d'une face pleine dans la couleur "teinte" *) 
let quad t =fun (F(a,b,c,d,_,teinte))->(
	set_color teinte;
	fill_poly [|(int_of_float(absc t.(a)),int_of_float(ord t.(a)));
			(int_of_float(absc t.(b)),int_of_float(ord t.(b)));
			(int_of_float(absc t.(c)),int_of_float(ord t.(c)));
			(int_of_float(absc t.(d)),int_of_float(ord t.(d)))|]
	);;

(* Prcalcul des cos et sin des angles  l'aide des formules de Tchebychev. 
L'argument est une fraction de 2P(en ralit un entier qui reprsente le 
diviseur) pour pouvoir reboucler au dbut du tableau lorsque l'on arrive 
 la fin *) 
let calcul_trigo portion_2pi =
	let pi=4.*.atan(1.)
in	let angle=2.*.pi/.(float_of_int portion_2pi)
in	let c1=cos angle and t=Array.make portion_2pi (0.,0.)
in	t.(0)<-(1.,0.); t.(1)<-(c1, sin angle);
	for c=2 to (portion_2pi-1) do
		t.(c) <- (	2.*.c1*.(fst t.(c-1))-.(fst t.(c-2)),
				2.*.c1*.(snd t.(c-1))-.(snd t.(c-2)) )
	done;
	t;;

(* Tableaux des cos et sin *) 
let trigo_theta = calcul_trigo 32;;
let trigo_phi   = calcul_trigo 128;;
let trigo_alpha = calcul_trigo 64;;

(* Cration d'un cube de c2l centr sur le point (x,y,z) orient selon les 
angles theta, phi et alpha *) 
let cree_cube (POri(x, y, z, theta, phi, alpha)) l =
	let ctheta=fst trigo_theta.(theta) 	and stheta=snd trigo_theta.(theta)
	and cphi=fst trigo_phi.(phi)	 		and sphi=snd trigo_phi.(phi)
	and calpha=fst trigo_alpha.(alpha) 	and salpha=snd trigo_alpha.(alpha)

in	(* Vecteurs d'orientation *) 
	let u= (P3(l*.ctheta*.cphi, -.l*.sphi, -.l*.stheta*.cphi))
	and v= (P3(l*.ctheta*.sphi*.calpha -. l*.stheta*.salpha,l*. cphi*.calpha, l*.(-.stheta*.sphi*.calpha -.ctheta *. salpha)))
	and w= (P3(l*.ctheta*.sphi*.salpha +. l*.stheta*.calpha,  l*.cphi*.salpha, l*.(-.stheta*.sphi*.salpha +. ctheta *. calpha)))

in 	(* Tableau des sommets et des normales, liste des faces *) 
	([|plus (plus (plus (P3(x, y, z)) u) v) w;
	   moins (plus (plus (P3(x, y, z)) u) v) w;
	   plus (moins (plus (P3(x, y, z)) u) v) w;
	   moins (moins (plus (P3(x, y, z)) u) v) w;
	   plus (plus (moins (P3(x, y, z)) u) v) w;
	   moins (plus (moins (P3(x, y, z)) u) v) w;
	   plus (moins (moins (P3(x, y, z)) u) v) w;
	   moins (moins (moins (P3(x, y, z)) u) v) w|],
	 [|u;v;w;neg u;neg v;neg w|],
	 [(F(0,1,3,2,0,rgb 192 0 0)); (F(4,5,7,6,3,rgb 192 0 0)); (F(0,1,5,4,1,rgb 192 0 0)); 
	  (F(2,3,7,6,4,rgb 192 0 0)); (F(0,2,6,4,2,rgb 192 0 0)); (F(1,3,7,5,5,rgb 192 0 0))])
;;

(* Recherche des faces visibles et calcul de la couleur en fonction de l'inclinaison 
par rapport au vecteur observateur *) 
let select (t_points,t_normales,l_faces) =
	let ombre c r=
		let red=int_of_float(float_of_int(c/65536)*.r)
		and green=int_of_float(float_of_int((c mod 65536)/256)*.r)
		and blue=int_of_float(float_of_int(c mod 256)*.r)
in 	rgb red green blue
in	let rec trouve = function
		|F(a,b,c,d,n,t)::q ->
						let r1= (prod_scal_norm (t_points.(a)) (t_normales.(n)))
						and r2= -.(prod_scal_norm (P3(1.,0.,0.)) (t_normales.(n)))
						in if r1<0. 
							then F(a,b,c,d,n,ombre t r2)::trouve q
							else trouve q
		|[] -> []
in	(t_points,trouve l_faces)
;;

(* Projection de tout le tableau de points 3D *) 	
let projete_ts_les_pts (t_pts,l_faces) = (Array.map proj t_pts, l_faces);;

(* Dessine toutes les faces *) 
let dessiner (t_pts,l_faces) = List.iter (quad t_pts) l_faces ;; 

(* Dtermine le plus petit rectangle contenant tous les points et dont les 
cts sont parallles  l'cran *) 
let map_vect_list f tab =
	let rec calcul result = function
		|0 -> f tab.(0)::result
		|n -> calcul ((f tab.(n))::result) (n-1)
in calcul [] (Array.length tab -1);;

let clip_rect (t_pts,_)=
let pair_of_list_of_vect_of_pt2d t = (map_vect_list absc t,map_vect_list ord t) 
and mini = function (a::b)-> (List.fold_left min a b)
	  |_-> failwith("Erreur dans MINI")
and maxi = function (a::b)-> (List.fold_left max a b)
	  |_-> failwith("Erreur dans MAXI")
in let tab_list = pair_of_list_of_vect_of_pt2d t_pts

in	let g=int_of_float (mini (fst tab_list))	(* gauche *)
	and d=int_of_float (maxi (fst tab_list))	(* droite *)
	and b=int_of_float (mini (snd tab_list))	(* bas *) 
	and h=int_of_float (maxi (snd tab_list))	(* haut *)
in	(g,b,d-g+1,h-b+1);; 				(* Les +1 corrigent l'erreur dau calcul entier *)

(* Une petite dmonstration !! *)
open_graph "";;
clear_graph();;

let uncurry f = function (a,b,c,d) -> f a b c d;;

let demo a=
let af =float_of_int a in
let projete=projete_ts_les_pts (select(cree_cube (POri(450.+.100.*.sin (af/.10.),100.*.cos (af/.10.),100.*.sin (af/.3.), a mod 32,a mod 128,a mod 64)) 100.))
in
(
dessiner (projete);
for a=1 to 2000 do
	if key_pressed() then failwith "Arrt utilisateur";
done;
set_color background;
uncurry fill_rect (clip_rect projete);
);;

let a=ref 0;;
while not(key_pressed()) do
	demo !a;
	a:=!a+1
done;;


